@@ -1,4 +1,4 @@
-/* This file is part of kdev-pg
+/* This file is part of kdev-pg-qt
    Copyright (C) 2005 Roberto Raggi <roberto@kdevelop.org>
 
    This library is free software; you can redistribute it and/or
@@ -39,7 +39,7 @@ private:
   QTextStream& strm;
 };
 
-void format(QTextStream &in, QTextStream& out);
+void format(QTextStream &in, const QString& oname);
 
 }
 
